"""
hudai.resources.person_quote
"""
from ..helpers.resource import Resource


class PersonQuoteResource(Resource):
    def __init__(self, client):
        Resource.__init__(self, client, base_path='/people/{person_id}/quotes')
        self.resource_name = 'PersonQuote'

    def list(self, person_id, article_id=None, term=None, page=None):
        query_params = self._set_limit_offset({'page': page})
        query_params['article_id'] = article_id
        query_params['term'] = term

        return self.http_get('/', params={'person_id': person_id},
                             query_params=query_params)

    def create(self, person_id, article_id, term, text):
        return self.http_post('/', params={'person_id': person_id},
                              data={
                                  'article_id': article_id,
                                  'term': term,
                                  'text': text
                              })

    def fetch(self, person_id, quote_id):
        return self.http_get('/{id}',
                             params={'person_id': person_id, 'id': quote_id})

    def delete(self, person_id, quote_id):
        return self.http_delete('/{id}',
                                params={'person_id': person_id, 'id': quote_id})



"""
hudai.resources.user_key_term
"""
from ..helpers.resource import Resource


class UserKeyTermResource(Resource):
    def __init__(self, client):
        Resource.__init__(self, client, base_path='/users/key-terms')
        self.resource_name = 'UserKeyTerm'

    def list(self, user_id, page=None):
        query_params = self._set_limit_offset({'page': page})
        query_params['user_id'] = user_id

        return self.http_get('/', query_params=query_params)

    def create(self, user_id, term):
        return self.http_post('/', data={'term': term, 'user_id': user_id})

    def fetch(self, user_id, term):
        return self.http_get('/{term}',
                             query_params={'user_id': user_id},
                             params={'term': term})

    def delete(self, user_id, term):
        return self.http_delete('/{term}',
                                query_params={'user_id': user_id},
                                params={'term': term})



"""
hudai.resources.system_task
"""
from datetime import datetime

from ..helpers.resource import Resource
from ..helpers.http_errors import api_404


class SystemTaskResource(Resource):
    def __init__(self, client):
        Resource.__init__(self, client, base_path='/system-tasks')
        self.resource_name = 'SystemTask'

    def list(self,
             page=None,
             started_after=None,
             started_before=None,
             completed=None,
             task_id=None):
        """
        Lists the existing system tasks, filtered by the given keyword args
        """
        return self._list(
            task_id=task_id,
            page=page,
            started_after=started_after,
            started_before=started_before,
            completed=completed)

    def create(self,
               task_id=None,
               attempts=None,
               started_at=None,
               completed_at=None):
        """
        Creates a new system task
        """
        return self._create(
            task_id=task_id,
            attempts=attempts,
            started_at=started_at,
            completed_at=completed_at)

    def fetch(self, entity_id):
        """
        Retrieves the task by its API id
        """
        return self._fetch(entity_id)

    def fetch_by_task_id(self, task_id):
        """
        Retrieves the task using its celery id
        """
        tasks = self._list(task_id=task_id).get('rows', [])
        return tasks[0] if tasks else api_404()

    def update(self,
               entity_id,
               attempts=None,
               started_at=None,
               completed_at=None):
        """
        Update the task's started_at or completed_at
        """
        return self._update(entity_id,
                            attempts=attempts,
                            started_at=started_at,
                            completed_at=completed_at)

    def mark_complete(self, entity_id, completed_at=datetime.now()):
        """
        Marks the task as complete (defaults to completed now)
        """
        return self._update(entity_id, completed_at=completed_at)

    def delete(self, entity_id):
        """
        Removes the task entry
        """
        return self._delete(entity_id)



"""HUD.ai Client Specs

Makes sure that the client surfaces the right resources and performs the
required token exchanges
"""

from datetime import datetime, timedelta
import pytest
import requests

from hudai.client import Client
from hudai.error import HudAiError
from hudai.resources import *


MOCK_CLIENT_ID = '7fe475ee-aae1-4ff2-82f4-ab48948edad0'
MOCK_CLIENT_SECRET = '42202699-d526-4505-9218-28e2af284e70'
MOCK_REDIRECT_URI = 'https://app.example.com/auth/callbacks/hud-ai'

def test_initialization():
    client = Client(client_id=MOCK_CLIENT_ID)

    assert isinstance(client, Client)
    assert isinstance(client.article_highlights, ArticleHighlightsResource)
    assert isinstance(client.article_key_terms, ArticleKeyTermResource)
    assert isinstance(client.articles, ArticleResource)
    assert isinstance(client.company_key_terms, CompanyKeyTermResource)
    assert isinstance(client.companies, CompanyResource)
    assert isinstance(client.domains, DomainResource)
    assert isinstance(client.key_terms, KeyTermResource)
    assert isinstance(client.people, PersonResource)
    assert isinstance(client.person_key_terms, PersonKeyTermResource)
    assert isinstance(client.person_quotes, PersonQuoteResource)
    assert isinstance(client.relevant_articles, RelevantArticlesResource)
    assert isinstance(client.system_events, SystemEventResource)
    assert isinstance(client.system_tasks, SystemTaskResource)
    assert isinstance(client.text_corpora, TextCorpusResource)
    assert isinstance(client.user_companies, UserCompanyResource)
    assert isinstance(client.user_contacts, UserContactResource)
    assert isinstance(client.user_digest_subscriptions, UserDigestSubscriptionResource)
    assert isinstance(client.user_key_terms, UserKeyTermResource)
    assert isinstance(client.users, UserResource)

def test_deprecated_methods():
    client = Client(client_id=MOCK_CLIENT_ID)

    assert isinstance(client.article_highlights, ArticleHighlightsResource)
    assert isinstance(client.article_key_term, ArticleKeyTermResource)
    assert isinstance(client.article, ArticleResource)
    assert isinstance(client.company_key_term, CompanyKeyTermResource)
    assert isinstance(client.company, CompanyResource)
    assert isinstance(client.domain, DomainResource)
    assert isinstance(client.key_term, KeyTermResource)
    assert isinstance(client.system_event, SystemEventResource)
    assert isinstance(client.system_task, SystemTaskResource)
    assert isinstance(client.text_corpus, TextCorpusResource)
    assert isinstance(client.user_company, UserCompanyResource)
    assert isinstance(client.user_key_term, UserKeyTermResource)
    assert isinstance(client.user, UserResource)


def test_get_authorize_uri_without_redirect_uri():
    client = Client(client_id=MOCK_CLIENT_ID)

    with pytest.raises(HudAiError):
        client.get_authorize_uri()


def test_get_authorize_uri():
    redirect_uri = 'https://app.example.com/auth/callbacks/hud-ai'
    client = Client(client_id=MOCK_CLIENT_ID, redirect_uri=redirect_uri)

    expected_uri = (
        'https://accounts.hud.ai/oauth2/authorize' +
        '?response_type=code' +
        '&client_id={client_id}' +
        '&redirect_uri={redirect_uri}'
    ).format(client_id=MOCK_CLIENT_ID, redirect_uri=redirect_uri)

    assert client.get_authorize_uri() == expected_uri


def test_refresh_tokens_valid_token(mocker):
    client = Client(client_id=MOCK_CLIENT_ID)
    client.token_expires_at = datetime.now() + timedelta(hours=1)

    mocker.spy(client, '_refresh_tokens')

    client.refresh_tokens()

    assert client._refresh_tokens.call_count == 0


def test_refresh_tokens_expired_token(mocker):
    client = Client(client_id=MOCK_CLIENT_ID, client_secret=MOCK_CLIENT_SECRET)

    mock_refresh_token = 'mock-refresh-token'

    client.token_expires_at = datetime.now() - timedelta(hours=1)
    client.refresh_token = mock_refresh_token

    mock_response = mocker.Mock()
    mock_json = {
        'access_token': 'new-token',
        'refresh_token': 'new-refresh-token',
        'expires_in': 100000
    }
    mock_response.json.return_value = mock_json

    mocker.patch('requests.post', return_value=mock_response)

    client.refresh_tokens()

    request_args, request_kwargs = requests.post.call_args

    assert request_args[0] == 'https://accounts.hud.ai/oauth2/token'
    assert request_kwargs['json'] == {
        'client_id': MOCK_CLIENT_ID,
        'client_secret': MOCK_CLIENT_SECRET,
        'grant_type': 'refresh_grant',
        'refresh_token': mock_refresh_token
    }
    assert client.access_token == mock_json['access_token']
    assert client.refresh_token == mock_json['refresh_token']

    projected_expiration = datetime.now() + \
        timedelta(milliseconds=mock_json['expires_in'])

    assert (
        (projected_expiration - timedelta(seconds=1))
        < client.token_expires_at <
        (projected_expiration + timedelta(seconds=1))
    )


def test_refresh_tokens_auth_code(mocker):
    client = Client(
        client_id=MOCK_CLIENT_ID,
        client_secret=MOCK_CLIENT_SECRET,
        redirect_uri=MOCK_REDIRECT_URI
    )

    mock_auth_code = 'mock-auth-code'

    client.set_auth_code(mock_auth_code)

    mock_response = mocker.Mock()
    mock_json = {
        'access_token': 'new-token',
        'refresh_token': 'new-refresh-token',
        'expires_in': 100000
    }
    mock_response.json.return_value = mock_json

    mocker.patch('requests.post', return_value=mock_response)

    client.refresh_tokens()

    request_args, request_kwargs = requests.post.call_args

    assert request_args[0] == 'https://accounts.hud.ai/oauth2/token'
    assert request_kwargs['json'] == {
        'client_id': MOCK_CLIENT_ID,
        'client_secret': MOCK_CLIENT_SECRET,
        'grant_type': 'authorization_code',
        'code': mock_auth_code,
        'redirect_uri': MOCK_REDIRECT_URI
    }
    assert client.access_token == mock_json['access_token']
    assert client.refresh_token == mock_json['refresh_token']

    projected_expiration = datetime.now() + \
        timedelta(milliseconds=mock_json['expires_in'])

    assert (
        (projected_expiration - timedelta(seconds=1))
        < client.token_expires_at <
        (projected_expiration + timedelta(seconds=1))
    )


def test_refresh_tokens_client_credentials(mocker):
    client = Client(client_id=MOCK_CLIENT_ID, client_secret=MOCK_CLIENT_SECRET)

    mock_response = mocker.Mock()
    mock_json = {
        'access_token': 'new-token',
        'refresh_token': 'new-refresh-token',
        'expires_in': 100000
    }
    mock_response.json.return_value = mock_json

    mocker.patch('requests.post', return_value=mock_response)

    client.refresh_tokens()

    request_args, request_kwargs = requests.post.call_args

    assert request_args[0] == 'https://accounts.hud.ai/oauth2/token'
    assert request_kwargs['json'] == {
        'client_id': MOCK_CLIENT_ID,
        'client_secret': MOCK_CLIENT_SECRET,
        'grant_type': 'client_credentials'
    }
    assert client.access_token == mock_json['access_token']
    assert client.refresh_token == mock_json['refresh_token']

    projected_expiration = datetime.now() + \
        timedelta(milliseconds=mock_json['expires_in'])

    assert (
        (projected_expiration - timedelta(seconds=1))
        < client.token_expires_at <
        (projected_expiration + timedelta(seconds=1))
    )



"""HUD.ai Resource Specs

Resources define collections of actions that can be performed on the same object
set (e.g. a User), and include a set of common (inheritable) protected methods
around CRUD actions.
"""
import pytest
from pytest_mock import mocker

from hudai import Client as HudAi
from hudai.helpers.http_client import HttpClient
from hudai.helpers.resource import Resource

HUD_CLIENT = HudAi(client_id='mock-client-id')
BASE_URL = 'https://api.hud.ai/v1'
HTTP_CLIENT = HttpClient(HUD_CLIENT, BASE_URL)

def test_standard_http_verbs_available():
    resource = Resource(HTTP_CLIENT)

    assert callable(resource.http_get)
    assert callable(resource.http_post)
    assert callable(resource.http_put)
    assert callable(resource.http_patch)
    assert callable(resource.http_delete)


@pytest.mark.parametrize('http_verb', [('get'), ('post'), ('put'), ('patch'), ('delete')])
def test_setting_base_path(mocker, http_verb):
    """
    Ensure that the `base_path` given when the resource is initialized is
    injected into all requests performed by that resource
    """
    method_name = "http_{}".format(http_verb)
    mocker.patch.object(HTTP_CLIENT, method_name, autospec=True)
    resource = Resource(HTTP_CLIENT, base_path='/test')

    function_under_test = getattr(resource, method_name)
    client_function = getattr(HTTP_CLIENT, method_name)

    function_under_test('/foo/bar', params={})

    client_function.assert_called_once_with('/test/foo/bar')


@pytest.mark.parametrize('http_verb', [('get'), ('post'), ('put'), ('patch'), ('delete')])
def test_parameter_injection(mocker, http_verb):
    """
    Ensure that the resource injects `params` into the URL when there is a
    matching keyword
    """
    method_name = "http_{}".format(http_verb)
    mocker.patch.object(HTTP_CLIENT, method_name, autospec=True)
    resource = Resource(HTTP_CLIENT)

    function_under_test = getattr(resource, method_name)
    client_function = getattr(HTTP_CLIENT, method_name)

    function_under_test('/test/{replace_me}/path',
                        params={'replace_me': 'replaced'})

    client_function.assert_called_once_with('/test/replaced/path')


def test__list(mocker):
    """
    Ensure that the inheritable method `_list` acts as expected
    (performs GET with params)
    """
    mocker.patch.object(HTTP_CLIENT, 'http_get', autospec=True)
    resource = Resource(HTTP_CLIENT, base_path='/mock-resource')

    resource._list(foo='bar')

    HTTP_CLIENT.http_get.assert_called_once_with('/mock-resource/',
                                            query_params={
                                                'foo': 'bar',
                                                'limit': 50,
                                                'offset': 0
                                            })


def test__create(mocker):
    """
    Ensure that the inheritable method `_create` acts as expected
    (performs POST with params)
    """
    mocker.patch.object(HTTP_CLIENT, 'http_post', autospec=True)
    resource = Resource(HTTP_CLIENT, base_path='/mock-resource')

    resource._create(foo='bar')

    HTTP_CLIENT.http_post.assert_called_once_with('/mock-resource/',
                                             data={'foo': 'bar'})


def test__fetch(mocker):
    """
    Ensure that the inheritable method `_fetch` acts as expected
    (performs GET action for given ID)
    """
    mocker.patch.object(HTTP_CLIENT, 'http_get', autospec=True)
    resource = Resource(HTTP_CLIENT, base_path='/mock-resource')

    resource._fetch('fake-uuid')

    HTTP_CLIENT.http_get.assert_called_once_with('/mock-resource/fake-uuid')


def test__update(mocker):
    """
    Ensure that the inheritable method `_update` acts as expected
    (performs PUT action for given ID, passing data)
    """
    mocker.patch.object(HTTP_CLIENT, 'http_put', autospec=True)
    resource = Resource(HTTP_CLIENT, base_path='/mock-resource')

    resource._update('fake-uuid', foo='bar')

    HTTP_CLIENT.http_put.assert_called_once_with('/mock-resource/fake-uuid',
                                            data={'foo': 'bar'})


def test__delete(mocker):
    """
    Ensure that the inheritable method `_delete` acts as expected
    (performs DELETE action for given ID)
    """
    mocker.patch.object(HTTP_CLIENT, 'http_delete', autospec=True)
    resource = Resource(HTTP_CLIENT, base_path='/mock-resource')

    resource._delete('fake-uuid')

    HTTP_CLIENT.http_delete.assert_called_once_with('/mock-resource/fake-uuid')



"""hudai.helpers.http_errors

Mock HTTP responses used when the wrapper does magic on behalf of the API for
convenience
"""

def api_404():
    """
    Returns standard 404 object
    """
    return {
        'message': 'Not Found',
        'status': 404,
        'type': 'not_found_error'
    }
